/*-------------------------------------------------------------------

    STEPD_GLA.C

    Ismo Krkkinen

    Stepwise search algorithm in decreasing direction using GLA.

-------------------------------------------------------------------*/

#define VersionNumber  "v0.1.1"
#define LastUpdated    "22.2.2002" /* iak */

#include "stepd_gla.h"
#include "cb_util.h"
#include "gla.h"
#include "cb.h"
#include "interfc.h"
#include <string.h>
#include "random.h"
#include <stdlib.h>
#include "memctrl.h"

static void PrintOutput(
    int Level, int M, int IterCount, float Best, float Current)
{
    switch (Level) {
    case 0: break;
    case 1:
	if (Best == Current) PrintMessage("M = %i, best = %#.6g.\n", M, Best);
	break;
    case 2:
    case 3:
	PrintMessage("M = %i, best = %#.6g, current = %#.6g, iter = %i.\n",
	    M, Best, Current, IterCount);
	break;
    default:
	PrintMessage("M = %i, best = %#.6g.\n", M, Best);
	break;
    }
}

static void PrintTrace(int Print, int ID, int Groups, int Iter, float Error)
{
    if (Print) ErrorMessage("%i\t%i\t%i\t%e\n", ID, Groups, Iter, Error);
}


typedef struct {
    int Number;
    long Key;
} PermutationData;

static int perm_asc(const void* a, const void* b) {
    const PermutationData* A = (PermutationData*) a;
    const PermutationData* B = (PermutationData*) b;
    if (A->Key > B->Key) return 1;
    if (A->Key < B->Key) return -1;
    return 0;
}

static void Permute(PermutationData* Permutation, int Count) {
    int k;
    for (k = 0; k < Count; k++) {
        Permutation[k].Number = k;
        Permutation[k].Key = CombTaus();
    }
    qsort(Permutation, Count, sizeof(PermutationData), &perm_asc);
}


static int PickBestCandidateDecr(TRAININGSET* TS,
    CODEBOOK* CB, PARTITIONING* PA,
    CriterionInfo* CI, float* Error, int Solutions)
{
    CODEBOOK CandCB, BestCB;
    PARTITIONING CandPA, BestPA;
    float CandError;
    int k, IterationCount = 0;
    PermutationData* Perm = allocate(sizeof(PermutationData) * BookSize(CB));
    Permute(Perm, BookSize(CB));
    *Error = 1e30;
    CreateNewCodebook(&CandCB, BookSize(CB) - 1, TS);
    CreateNewCodebook(&BestCB, BookSize(CB) - 1, TS);
    CreateNewPartitioning(&CandPA, TS, BookSize(CB) - 1);
    CreateNewPartitioning(&BestPA, TS, BookSize(CB) - 1);
    if (Solutions > BookSize(CB)) Solutions = BookSize(CB);
    for (k = 0; k < Solutions; ++k) {
        CopyCodebookOmitting(&CandCB, CB, Perm[k].Number);
        IterationCount +=
            GeneralizedLloydAlgorithm(TS, &CandCB, &CandPA, 0, CI);
        CandError = ciEvaluate(CI, &CandCB, &CandPA);
        if (CandError < *Error) {
            *Error = CandError;
            CopyCodebook(&CandCB, &BestCB);
            CopyPartitioning(&CandPA, &BestPA);
        }
    }
    DecreaseCodebookSize(CB, BookSize(&BestCB));
    CopyCodebook(&BestCB, CB);
    JoinPartitions(TS, PA, 0, PartitionCount(PA) - 1);
    CopyPartitioning(&BestPA, PA);
    FreeCodebook(&CandCB);
    FreeCodebook(&BestCB);
    FreePartitioning(&CandPA);
    FreePartitioning(&BestPA);
    deallocate(Perm);
    return IterationCount;
}

int StepwiseDecrGLA(TRAININGSET* TSIn, CODEBOOK* CBOut, PARTITIONING* PAOut,
    int MinGroups, int MaxGroups, CriterionInfo* CI, float* Error,
    int Solutions, int Output, int Trace, int TraceID)
{
    int IterCount, k, GLA_Output;
    float Curr;
    CODEBOOK CurrCB;
    PARTITIONING CurrPA;
    if (Output > 2) GLA_Output = Output - 2;
    else GLA_Output = 0;
    CreateNewCodebook(CBOut, MaxGroups, TSIn);
    AddGenerationMethod(CBOut, "Stepwise GLA");
    CreateNewPartitioning(PAOut, TSIn, MaxGroups);
    IterCount = RandomizedGeneralizedLloydAlgorithm(TSIn, CBOut, PAOut,
        CI, Error, Solutions, 200, GLA_Output, Trace, TraceID);
    PrintOutput(Output, MaxGroups, IterCount, *Error, *Error);
    CreateNewCodebook(&CurrCB, MaxGroups, TSIn);
    CreateNewPartitioning(&CurrPA, TSIn, MaxGroups);
    CopyCodebook(CBOut, &CurrCB);
    CopyPartitioning(PAOut, &CurrPA);
    for (k = MaxGroups - 1; k >= MinGroups; --k) {
        IterCount = PickBestCandidateDecr(
            TSIn, &CurrCB, &CurrPA, CI, &Curr, Solutions);
        if (*Error > Curr) {
            DecreaseCodebookSize(CBOut, BookSize(&CurrCB));
            CopyCodebook(&CurrCB, CBOut);
            while (k < PartitionCount(PAOut))
                JoinPartitions(TSIn, PAOut, 0, PartitionCount(PAOut) - 1);
            CopyPartitioning(&CurrPA, PAOut);
            *Error = Curr;
        }
	PrintOutput(Output, k, IterCount, *Error, Curr);
        PrintTrace(Trace, TraceID, k, IterCount, Curr);
    }
    FreeCodebook(&CurrCB);
    FreePartitioning(&CurrPA);
    return 0;
}

