/*-------------------------------------------------------------------

    STEPD_RLS.C

    Ismo Krkkinen

    Stepwise search algorithm in decreasing direction using RLS.

-------------------------------------------------------------------*/

#define VersionNumber  "v0.1.1"
#define LastUpdated    "8.1.2003" /* iak */

#include "stepd_rls.h"
#include "cb_util.h"
#include "rls.h"
#include "cb.h"
#include "interfc.h"
#include <string.h>
#include "random.h"
#include <stdlib.h>
#include "memctrl.h"

static void PrintOutput(
    int Level, int M, int IterCount, float Best, float Current)
{
    switch (Level) {
    case 0: break;
    case 1:
	if (Best == Current) PrintMessage("M = %i, best = %#.6g.\n", M, Best);
	break;
    case 2:
    case 3:
	PrintMessage("M = %i, best = %#.6g, current = %#.6g, iter = %i.\n",
	    M, Best, Current, IterCount);
	break;
    default:
	PrintMessage("M = %i, best = %#.6g.\n", M, Best);
	break;
    }
}

static void PrintTrace(int Print, int ID, int Groups, int Iter, float Error) {
    if (Print) ErrorMessage("%i\t%i\t%i\t%e\n", ID, Groups, Iter, Error);
}


typedef struct {
    int Number;
    long Key;
} PermutationData;

static int perm_asc(const void* a, const void* b) {
    const PermutationData* A = (PermutationData*) a;
    const PermutationData* B = (PermutationData*) b;
    if (A->Key > B->Key) return 1;
    if (A->Key < B->Key) return -1;
    return 0;
}

static void Permute(PermutationData* Permutation, int Count) {
    int k;
    for (k = 0; k < Count; k++) {
        Permutation[k].Number = k;
        Permutation[k].Key = CombTaus();
    }
    qsort(Permutation, Count, sizeof(PermutationData), &perm_asc);
}

static int PickBestCandidateDecr(TRAININGSET* TS,
    CODEBOOK* CB, PARTITIONING* PA,
    CriterionInfo* CI, float* Error, int Solutions,
    Iterations* Iter, int GLAIterations, int SwapMethod, int Output)
{
    CODEBOOK CandCB, BestCB;
    PARTITIONING CandPA, BestPA;
    float CandError;
    int k, IterCount = 0;
    int RLSOutput;
    int* Changed;
    PermutationData* Perm = allocate(sizeof(PermutationData) * BookSize(CB));
    Permute(Perm, BookSize(CB));
    *Error = 1e30;
    if (Output > 2) RLSOutput = Output - 2;
    else RLSOutput = 0;
    CreateNewCodebook(&CandCB, BookSize(CB) - 1, TS);
    CreateNewCodebook(&BestCB, BookSize(CB) - 1, TS);
    CreateNewPartitioning(&CandPA, TS, BookSize(CB) - 1);
    CreateNewPartitioning(&BestPA, TS, BookSize(CB) - 1);
    Changed = allocate(sizeof(int) * (BookSize(CB) - 1));
    if (Solutions > BookSize(CB)) Solutions = BookSize(CB);
    for (k = 0; k < Solutions; ++k) {
        CopyCodebookOmitting(&CandCB, CB, Perm[k].Number);
        memset(Changed, 1, sizeof(int) * (BookSize(CB) - 1));
        ciPartitionOptimally(CI, &CandCB, &CandPA, Changed);
        FixCodebookFrequencies(&CandCB, &CandPA);
        itReset(Iter);
        RandomizedLocalSearch(TS, &CandCB, &CandPA,
            CI, &CandError, Iter, GLAIterations, SwapMethod, RLSOutput, 0, 0);
        IterCount += itCount(Iter);
        if (CandError < *Error) {
            *Error = CandError;
            CopyCodebook(&CandCB, &BestCB);
            CopyPartitioning(&CandPA, &BestPA);
        }
        if (Output <= 2) PrintOutput(
            Output, BookSize(&CandCB), IterCount, *Error, CandError);
    }
    deallocate(Changed);
    DecreaseCodebookSize(CB, BookSize(&BestCB));
    CopyCodebook(&BestCB, CB);
    JoinPartitions(TS, PA, 0, PartitionCount(PA) - 1);
    CopyPartitioning(&BestPA, PA);
    FreePartitioning(&BestPA);
    FreePartitioning(&CandPA);
    FreeCodebook(&BestCB);
    FreeCodebook(&CandCB);
    deallocate(Perm);
    return IterCount;
}

int StepwiseDecrRLS(TRAININGSET* TSIn, CODEBOOK* CBOut, PARTITIONING* PAOut,
    int MinGroups, int MaxGroups, CriterionInfo* CI, float* Error,
    int Solutions, Iterations* Iter, int GLAIterations, int SwapMethod,
    int Output, int Trace, int TraceID)
{
    int k, RLS_Output, IterCount;
    float Curr;
    CODEBOOK CurrCB;
    PARTITIONING CurrPA;
    int* Changed;
    if (Output > 2) RLS_Output = Output - 2;
    else RLS_Output = 0;
    CreateNewCodebook(CBOut, MaxGroups, TSIn);
    AddGenerationMethod(CBOut, "Stepwise RLS");
    GenerateRandomCodebook(TSIn, CBOut);
    CreateNewPartitioning(PAOut, TSIn, MaxGroups);
    Changed = allocate(sizeof(int) * BookSize(CBOut));
    memset(Changed, 1, sizeof(int) * (BookSize(CBOut) - 1));
    ciPartitionOptimally(CI, CBOut, PAOut, Changed);
    deallocate(Changed);
    FixCodebookFrequencies(CBOut, PAOut);
    RandomizedLocalSearch(TSIn, CBOut, PAOut, CI, Error,
        Iter, GLAIterations, SwapMethod, RLS_Output, 0, 0);
    IterCount = itCount(Iter);
    PrintOutput(Output, MaxGroups, IterCount, *Error, *Error);
    PrintTrace(Trace, TraceID, BookSize(CBOut), IterCount, *Error);
    CreateNewCodebook(&CurrCB, MaxGroups, TSIn);
    CreateNewPartitioning(&CurrPA, TSIn, MaxGroups);
    CopyCodebook(CBOut, &CurrCB);
    CopyPartitioning(PAOut, &CurrPA);
    for (k = MaxGroups - 1; k >= MinGroups; --k) {
        IterCount = PickBestCandidateDecr(TSIn, &CurrCB, &CurrPA,
            CI, &Curr, Solutions, Iter, GLAIterations, SwapMethod, RLS_Output);
        if (*Error > Curr) {
            DecreaseCodebookSize(CBOut, BookSize(&CurrCB));
            CopyCodebook(&CurrCB, CBOut);
            while (k < PartitionCount(PAOut))
                JoinPartitions(TSIn, PAOut, 0, PartitionCount(PAOut) - 1);
            CopyPartitioning(&CurrPA, PAOut);
            *Error = Curr;
        }
	PrintOutput(Output, k, IterCount, *Error, Curr);
        PrintTrace(Trace, TraceID, k, IterCount, Curr);
    }
    FreeCodebook(&CurrCB);
    FreePartitioning(&CurrPA);
    return 0;
}

